/*
  Copyright 2005 Damien Stehl.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  This program implements ideas from the paper "Floating-point LLL Revisited", 
  by Phong Nguyen and Damien Stehl, in the Proceedings of Eurocrypt'2005, 
  Springer-Verlag; and was partly inspired by Shoup's NTL library: 
  http://www.shoup.net/ntl/

  It is part of fplll-1.2, downloadable at:
  http://www.loria.fr/~stehle/fplll-1.2.tar.gz 

*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <ctype.h>
#include <limits.h>
#include "gmp.h"

#ifndef NOT_USING_MPFR
#define USING_MPFR
#endif

#ifdef USING_MPFR
#include "mpfr.h"
#endif

#if (LONG_MAX==2147483647L)
#define CPU_SIZE_1 31
#define MAX_LONG 0x1p31
#else
#define CPU_SIZE_1 53
#define MAX_LONG 0x1p53
#endif

#ifdef FORCE_LONGDOUBLE
#define double long double
#define sqrt sqrtl
#endif

#ifndef ETA
#define ETA 0.6
#endif
#ifndef DELTA
#define DELTA 0.8
#endif

gmp_randstate_t state;

#include "mat.h"
#include "basicfast.h"
#include "fast.h"

double halfplus, onedothalfplus, ctt;


// #define VERBOSE
// #define TRIANGULAR 
// #define DEBUG 

#ifdef DEBUG
#define LOOPS_BABAI 5
#endif

#ifdef TRIANGULAR
#ifndef SHIFT 
#define SHIFT 0
#endif
#endif

#ifdef VERBOSE
int
cputime ()
{
  struct rusage rus;

  getrusage (RUSAGE_SELF, &rus);
  return rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_usec / 1000;
}
#endif




/* computes the prec MSBs of the inner product of the approximate 
   vec1 and vec2. */


double
fpScalarProduct (double *vec1, double *vec2, int n)
{
  int i;
  double sum;

  sum = vec1[1] * vec2[1];
  for (i=2; i<=n; i++)
    sum += vec1[i] * vec2[i];

  return sum;
} 


double 
fpNorm (double *vec, int n)
{
  int i;
  double sum;

  sum = vec[1] * vec[1];
  for (i = 2 ; i <= n ; i++)
    sum += vec[i]*vec[i];

  return sum;

} 


/***********************************/
/* ******************************* */
/* Babai's Nearest Plane algorithm */
/* ******************************* */
/***********************************/

/* Size-reduces b_kappa using mu_ij and r_ij for j<=i <kappa
updates B (kappa)
computes mu_kappaj, r_kappaj for j<=kappa, and s(kappa) 
The algorithm is the iterative Babai algorithm of the paper
*/


void 
Babai (int kappa, mat_ZZ *B, double **mu, double **r, double *s, 
       double **appB, int *expo, double **appSP, 
       int a, int zeros, int kappamax, int n, mpz_t ztmp)
{
  int i, j, k, test, aa, exponent;
  signed long xx;
  double tmp, rtmp;
#ifdef DEBUG
  int loops=0;
#endif 


  if (a > zeros)
    aa = a;
  else
    aa = zeros+1;
  
#ifdef DEBUG
  printf("\nappSP: \n");
  Print_matf (appSP, expo, kappamax, kappamax);
  printf("\nr: \n");
  Print_matf (r, expo, kappamax, kappamax);
  printf("\n          STARTING BABAI WITH k=%d\n", kappa);
  printf("\nappB: \n");
  Print_matf (appB, expo, kappamax, n);
  loops = 0;
  printf("\nmu: \n");
  Print_matf (mu, expo, kappamax, kappamax);
  printf ("\n\na is %d, zeros is %d, aa is %d\n", a, zeros, aa);
  Print_mat (B->coeff, kappamax, n);
#endif


  do
    {
      test=0;
      
#ifdef DEBUG     
      if (loops++ > LOOPS_BABAI) 
	{
	  printf("INFINITE LOOP?\n"); 
	  exit(0);
	}
#endif 
      
      /* ************************************** */
      /* Step2: compute the GSO for stage kappa */
      /* ************************************** */
      
      for (j=aa; j<kappa; j++)
	{	  
	  if (appSP[kappa][j]!=appSP[kappa][j])
	    {
	      appSP[kappa][j] = fpScalarProduct (appB[kappa], appB[j], n);
	    }
	  
#ifdef DEBUG
	  if (loops<=LOOPS_BABAI)
	    {
	      printf("\n          j is %d\n", j);	  
#ifdef FORCE_LONGDOUBLE
	      printf("          appSP[%d][%d] is: %1.5Le", kappa, j, appSP[kappa][j]); 
#else
	      printf("          appSP[%d][%d] is: %E", kappa, j, appSP[kappa][j]); 
#endif
	      printf(", which approximates ");
	      ScalarProduct (ztmp, B->coeff[kappa], B->coeff[j], n);
	      printf("\n");
	      Print_mat (B->coeff, kappa, n);
	      mpz_out_str(stdout, 10, ztmp); 
	      printf("\n          Norm of B[%d]^2: ", j);
	      ScalarProduct (ztmp, B->coeff[j], B->coeff[j], n);
	      mpz_out_str(stdout, 10, ztmp);  
	      printf("\n          Norm of B[%d]^2: ",kappa);
	      ScalarProduct (ztmp, B->coeff[kappa], B->coeff[kappa], n);
	      mpz_out_str(stdout, 10, ztmp); 
	    }
#endif
	  
	  if (j > zeros+2)
	    {
	      tmp = mu[j][zeros+1] * r[kappa][zeros+1];
	      rtmp = appSP[kappa][j] - tmp;
	      for (k=zeros+2; k<j-1; k++)
		{
		  tmp = mu[j][k] * r[kappa][k];
		  rtmp = rtmp - tmp;
		}
	      tmp = mu[j][j-1] * r[kappa][j-1];
	      r[kappa][j] = rtmp - tmp;
	    }
	  else if (j==zeros+2)
	    {
	      tmp = mu[j][zeros+1] * r[kappa][zeros+1];
	      r[kappa][j] = appSP[kappa][j] - tmp;
	    }
	  else r[kappa][j] = appSP[kappa][j];

	  mu[kappa][j] = r[kappa][j] / r[j][j];
	}


      

#ifdef DEBUG
      if (loops <=LOOPS_BABAI)
	{
	  printf("\nmu :\n");
	  Print_matf (mu, expo, kappa, kappa);
	  printf("\nr :\n");
	  Print_matf (r, expo, kappa, kappa);
	}
#endif
      
      /* **************************** */
      /* Step3--5: compute the X_j's  */
      /* **************************** */
      
      for (j=kappa-1; j>zeros; j--)
	{

	  /* test of the relaxed size-reduction condition */
	  tmp = fabs (mu[kappa][j]);
	  tmp = ldexp (tmp, expo[kappa]-expo[j]);


#ifdef DEBUG
#ifdef FORCE_LONGDOUBLE
	  if (loops<=LOOPS_BABAI) printf( "tmp is: %1.5Le\n", tmp); 
#else
	  if (loops<=LOOPS_BABAI) printf( "tmp is: %E\n", tmp); 
#endif
#endif
	  
	  if (tmp > halfplus) 
	    {
	      test = 1; 
	      exponent = expo[j] - expo[kappa];
	      
	      /* we consider separately the cases X = +-1 */     
	      if (tmp <= onedothalfplus)   
		{
#ifdef DEBUG
		  printf(" X is pm1\n");
#endif
		  
		  if ( mu[kappa][j] >=0 )   /* in this case, X is 1 */
		    {
		      for (k=zeros+1; k<j; k++)
			{
			  tmp = ldexp (mu[j][k], exponent);
			  mu[kappa][k] =  mu[kappa][k] - tmp; 
			}
		      
		      for (i=1; i<=n; i++)
			mpz_sub (B->coeff[kappa][i], 
				 B->coeff[kappa][i], 
				 B->coeff[j][i]);	  
		    }
		  
		  else          /* otherwise X is -1 */ 
		    {
		      for (k=zeros+1; k<j; k++)
			{
			  tmp = ldexp (mu[j][k], exponent);
			  mu[kappa][k] = mu[kappa][k] + tmp;
			}
		      
		      for (i=1; i<=n; i++)
			mpz_add (B->coeff[kappa][i], 
				 B->coeff[kappa][i], 
				 B->coeff[j][i]);
		    }
		}
	      
	      else   /* we must have |X| >= 2 */
		{
		  
		  tmp = ldexp (mu[kappa][j] , -exponent);
		  
	          if ((tmp < (double) MAX_LONG)
		      &&(tmp > (double) -MAX_LONG))  
		    {
		      tmp = rint (tmp);
		      
		      for (k=zeros+1; k<j; k++)
			{
			  rtmp = tmp * mu[j][k];
			  rtmp = ldexp (rtmp, exponent);
			  mu[kappa][k] = mu[kappa][k] - rtmp;
			}
	
		      xx = (signed long int) tmp;
		      
#ifdef DEBUG
		      if (loops<=LOOPS_BABAI)
			{
			  printf("          xx[%d] is %ld\n", j, xx);
#ifdef FORCE_LONGDOUBLE
			  printf("          and tmp was %1.5Le\n", tmp);
#else
			  printf("          and tmp was %E\n", tmp);
#endif
			}
#endif

		      for (i=1; i<=n; i++)
			{
			  if (xx > 0)
			    {
			      mpz_submul_ui (B->coeff[kappa][i], B->coeff[j][i], (unsigned long int) xx);
			    }
			  else
			    {
			      mpz_addmul_ui (B->coeff[kappa][i], B->coeff[j][i], (unsigned long int) -xx);
			    }
			}
		    }
		  
		  else
		    {
		      tmp = frexp(mu[kappa][j], &exponent); 
		      tmp = tmp * MAX_LONG;
		      xx = (signed long int) tmp;
		      exponent = expo[kappa] - expo[j] + exponent - CPU_SIZE_1 ;

#ifdef DEBUG
#ifdef FORCE_LONGDOUBLE
		      printf("tmp is %1.5Le", tmp);
#else
		      printf("tmp is %E", tmp);
#endif
		      printf("\nand exponent is %d, and X is %ld\n", exponent, xx);
#endif

		      /* This case is extremely rare: never happened for me */
		      if (exponent <= 0) 
			{
#ifdef VERBOSE
			  fprintf(stderr, "Hello, this is the rare Babai case\n");
#endif
			  xx = xx << -exponent;
			  exponent = 0;
			  
			  for (i=1; i<=n; i++)
			    {
			      mpz_mul_si (ztmp, B->coeff[j][i], xx);
			      mpz_sub (B->coeff[kappa][i], 
				       B->coeff[kappa][i], ztmp);
			    }
			  for (k=zeros+1; k<j; k++)
			    {
			      rtmp = ((double) xx) * mu[j][k];
			      rtmp = ldexp (rtmp, expo[j]-expo[kappa]);
			      mu[kappa][k] = mu[kappa][k] - rtmp;
			    }
			  
			}
		      else
			{
			  for (i=1; i<=n; i++)
			    {
			      mpz_mul_2exp (ztmp, B->coeff[j][i], exponent); 
			      if (xx>0)
				{
				  mpz_submul_ui (B->coeff[kappa][i], ztmp, (unsigned long int) xx);
				}
			      else
				{
				  mpz_addmul_ui (B->coeff[kappa][i], ztmp, (unsigned long int) -xx);
				}
			    }
			  
			  for (k=zeros+1; k<j; k++)
			    {
			      rtmp = ((double) xx) * mu[j][k];
			      rtmp = ldexp (rtmp, exponent+expo[j]-expo[kappa]);
			      mu[kappa][k] = mu[kappa][k] - rtmp;
			    }
			}
		    }		  
		}
	    }
	}
      
      if (test)   /* Anything happened? */
	{
	  expo[kappa] = set_line (appB[kappa], B->coeff[kappa], n);
	  aa = zeros+1;
	  for (i=zeros+1; i<=kappa; i++) 
	    appSP[kappa][i] = 0.0/0.0;
	  for (i=kappa+1; i<=kappamax; i++) 
	    appSP[i][kappa] = 0.0/0.0;
	}
	  
#ifdef DEBUG
      if (loops<=LOOPS_BABAI)
	{
	  printf("          test is %d\n", test);
	  printf("\nmu: \n");
	  Print_matf (mu, expo, kappa, kappa);
	  printf("\nr: \n");
	  Print_matf (r, expo, kappa, kappa);
	}
#endif
      
    }
  while (test);



  if (appSP[kappa][kappa]!=appSP[kappa][kappa]) 
    {
      appSP[kappa][kappa] = fpNorm (appB[kappa], n);
    }
  s[zeros] = appSP[kappa][kappa];
  
  /* the last s[kappa-1]=r[kappa][kappa] is computed only if kappa increases */
  for (k=zeros+1; k<=kappa-2; k++)
    {
      tmp = mu[kappa][k] * r[kappa][k];
      s[k] = s[k-1] - tmp;
    }

#ifdef DEBUG
  printf("          Number of loops is %d\n", loops);
#endif

}





/* ****************** */
/* ****************** */
/* ****************** */
/* The LLL Algorithm  */
/* ****************** */
/* ****************** */
/* ****************** */


/* LLL-reduces the integer matrix B "in place" */


void 
LLL (mat_ZZ *B)
{
  int kappa, kappa2, d, n, i, j, zeros, kappamax;
  double **mu, **r, **appB, **appSP;
  double *s, *mutmp, *appBtmp, *appSPtmp;
  double tmp=0.0;
#ifdef FORCE_LONGDOUBLE
  double y;
#endif
  int *expo, *alpha;
  mpz_t ztmp;
  mpz_t *Btmp;
#ifdef VERBOSE
  int newkappa, start, lovasz, loops;
#endif
  
  ctt = DELTA;
  halfplus = ETA;

  n = B->NumRows;
  d = B->NumCols;


  alpha = (int *) malloc((d+1) * sizeof(int)); 
  expo = (int *) malloc((d+1) * sizeof(int)); 
  mpz_init (ztmp);

#ifdef DEBUG
  printf ("d = %d, n=%d\n", d, n);
#endif

#ifdef VERBOSE
#ifdef FORCE_LONGDOUBLE

  tmp = 1.0;
  for (j = 0; (y = tmp + 1.0) != tmp; j++, tmp = 2.0 * tmp);

  fprintf (stderr, 
	   "\nEntering LLL^2: LLL-reduction factors (%1.5Le, %1.5Le)\n", ctt, halfplus);  fprintf ( stderr, 
	    "Working precision set to %d bits.\n", j);
#else
  fprintf (stderr, 
	   "\nEntering LLL^2: LLL-reduction factors (%E, %E)\n", ctt, halfplus);

  fprintf ( stderr, 
	    "Working precision set to 53 bits.\n");
#endif
  fprintf (stderr, 
	   "No efficiency nor correctness guarantee here.\nIf you want to be sure, LLL_check your output\nbasis or use the guaranteed version.\n\n");
#endif


  mu = (double **) malloc ((d + 1) * sizeof (double*));
  for (i=0; i<=d; i++)
    mu[i] = (double*) malloc ((d+1) * sizeof (double));

  r = (double **) malloc ((d + 1) * sizeof (double*));
  for (i=0; i<=d; i++)
    r[i] = (double*) malloc ((d+1) * sizeof (double));

  appB = (double **) malloc ((d + 1) * sizeof (double*));
  for (i=0; i<=d; i++)
    appB[i] = (double*) malloc ((n+1) * sizeof (double));

  appSP = (double **) malloc ((d + 1) * sizeof (double*));
  for (i=0; i<=d; i++)
    appSP[i] = (double*) malloc ((d+1) * sizeof (double));

  s = (double *) malloc ((d + 1) * sizeof (double));
  appSPtmp = (double *) malloc ((d + 1) * sizeof (double));

  for (i=0; i<=d; i++)
    for (j=0; j<=d; j++)
      appSP[i][j] = 0.0/0.0;
  
  /* ************************** */
  /* Step1: Initialization Step */
  /* ************************** */     
  
#ifdef VERBOSE  
  start = cputime();
#endif
  
  for (i=1; i<=d; i++) {
    expo[i] = set_line (appB[i], B->coeff[i], n);
  }

#ifdef DEBUG  
  Print_matf (appB, expo, d, n);
  printf("Step 1 is over\n");
#endif  
  
  
  /* ********************************* */
  /* Step2: Initializing the main loop */
  /* ********************************* */   
  
  
#ifdef VERBOSE  
  newkappa = 1;
  loops = lovasz = 0;
#endif

  kappamax = 1;
  i = 1; 
  
  do
    appSP[i][i] = fpNorm (appB[i], n); 
  while ( (appSP[i][i] <=0.0)&&(++i <=d));
  zeros = i-1; /* all vectors B[i] with i <= zeros are zero vectors */
  kappa = i+1;
  
  if (zeros < d)  r[zeros+1][zeros+1] = appSP[zeros+1][zeros+1];

  for (i=zeros+1; i<=d; i++)
    alpha[i]=1;
  
#ifdef DEBUG
  Print_matf (appSP, expo, d, d);
  printf("Step 2 is over\n"); 
#endif  
  
  while (kappa <= d)
    {      

      if (kappa>kappamax) kappamax++;
      
#ifdef VERBOSE
      loops++;
      if (kappa>newkappa)
	{
	  newkappa++;
	  fprintf (stderr, 
		   "Discovering vector k = %d, iterations = %d ", 
		   kappa, loops);
	  fprintf(stderr, "cputime is %d ms\n", (cputime()-start));
	}
#endif
      
#ifdef DEBUG
      printf("alpha= ");
      for (i=zeros+1; i<=d; i++) 
	printf("%d ", alpha[i]);
      printf("\n");
      printf("entering the while loop with k=%d\n", kappa);
      Print_mat (B->coeff, d, n);
#endif

      /* ********************************** */
      /* Step3: Call to the Babai algorithm */
      /* ********************************** */   

#ifdef TRIANGULAR      
      if (kappamax + SHIFT <= n){	
	Babai (kappa, B, mu, r, s, appB, expo, appSP, 
	       alpha[kappa], zeros, kappamax, kappamax+SHIFT, ztmp);
      }
      else {
	Babai (kappa, B, mu, r, s, appB, expo, appSP, 
	       alpha[kappa], zeros, kappamax, n, ztmp);
      }
#else
      Babai (kappa, B, mu, r, s, appB, expo, appSP, 
	     alpha[kappa], zeros, kappamax, n, ztmp);
#endif

#ifdef DEBUG      
      printf("Step 3 is over\n");   
      Print_mat (B->coeff, kappamax, n);
      Print_matf (appSP, expo, kappamax, kappamax);
      printf("\nr: \n");
      Print_matf (r, expo, kappamax, kappamax);
#endif
      
      /* ************************************ */
      /* Step4: Success of Lovasz's condition */
      /* ************************************ */  
      /* ctt * r.coeff[kappa-1][kappa-1] <= s[kappa-2] ?? */
      

      tmp = r[kappa-1][kappa-1] * ctt;
      tmp = ldexp (tmp, 2*(expo[kappa-1]-expo[kappa]));


#ifdef DEBUG
#ifdef FORCE_LONGDOUBLE
      printf("s[%d] is %1.5Le\n", kappa-2, s[kappa-2]);
#else
      printf("s[%d] is %E\n", kappa-2, s[kappa-2]);
#endif
#endif
      
#ifdef VERBOSE
      lovasz++;
#endif

      if (tmp <= s[kappa-2]) 
	{
	  alpha[kappa] = kappa;
	  tmp = mu[kappa][kappa-1] * r[kappa][kappa-1];
	  r[kappa][kappa] = s[kappa-2] - tmp;
	  kappa++;
	} 
      else
	{

	  /* ******************************************* */
	  /* Step5: Find the right insertion index kappa */
          /* kappa2 remains the initial kappa            */
	  /* ******************************************* */  


	  kappa2 = kappa;
	  do
	    {
#ifdef VERBOSE
	      lovasz++;
#endif
	      kappa--;
	      tmp = r[kappa-1][kappa-1] * ctt;
	      tmp = ldexp (tmp, 2*(expo[kappa-1]-expo[kappa2]));
	    }
	  while ( (kappa >= zeros+2) && (s[kappa-2] <= tmp) );

#ifdef DEBUG
	  printf( "Index of insertion: %d \n", kappa);
	  printf("Step 5 is over\n");
	  printf("alpha= ");
	  for (i=1; i<=kappamax; i++) 
	    printf("%d ", alpha[i]);
	  printf("\n");
#endif
	  
	  for (i=kappa; i<kappa2; i++)
	    if (kappa <= alpha[i]) alpha[i] = kappa;

	  for (i=kappa2; i>kappa; i--)
	    alpha[i] = alpha[i-1];

	  for (i=kappa2+1; i<=kappamax; i++)
	    if (kappa < alpha[i]) alpha[i] = kappa;
	  
	  alpha[kappa] = kappa;

#ifdef DEBUG
	  printf("alpha= ");
	  for (i=1; i<=kappamax; i++) 
	    printf("%d ", alpha[i]);
	  printf("\n");
#endif

	  /* ****************************** */
	  /* Step6: Update the mu's and r's */
	  /* ****************************** */  
	  
	  mutmp = mu[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    mu[i] = mu[i-1];
	  mu[kappa] = mutmp;
	  
	  mutmp = r[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    r[i] = r[i-1];
	  r[kappa] = mutmp;

	  r[kappa][kappa] = s[kappa-1];
	  
#ifdef DEBUG 
	  printf("Step 6 is over\n");
#endif
	  
	  /* ************************ */
	  /* Step7: Update B and appB */
	  /* ************************ */  	  
	  
	  Btmp = B->coeff[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    B->coeff[i] = B->coeff[i-1];
	  B->coeff[kappa] = Btmp;
	  
	  appBtmp = appB[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    appB[i] = appB[i-1];
	  appB[kappa] = appBtmp;

	  j = expo[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    expo[i] = expo[i-1];
	  expo[kappa] = j;

	  
#ifdef DEBUG
	  Print_mat (B->coeff, d, n);
	  Print_matf (appB, expo, d, n);
	  printf("Step 7 is over\n");
#endif

	  /* *************************** */
	  /* Step8: Update appSP: tricky */
	  /* *************************** */  	 
	  
	  for (i=1; i<=kappa2; i++)
	    appSPtmp[i] = appSP[kappa2][i];

	  for (i=kappa2+1; i<=kappamax; i++)
	    appSPtmp[i] = appSP[i][kappa2];
	  
	  for (i=kappa2; i>kappa; i--)
	    {
	      for (j=1; j<kappa; j++)
		appSP[i][j] = appSP[i-1][j];
	      
	      appSP[i][kappa] = appSPtmp[i-1];
	      
	      for (j=kappa+1; j<=i; j++)
		appSP[i][j] = appSP[i-1][j-1];

	      for (j=kappa2+1; j<=kappamax; j++)
		appSP[j][i] = appSP[j][i-1];     
	    }
	  
	  for (i=1; i<kappa; i++)
	    appSP[kappa][i] = appSPtmp[i];

	  appSP[kappa][kappa] = appSPtmp[kappa2];

	  for (i=kappa2+1; i<=kappamax; i++)
	    appSP[i][kappa] = appSPtmp[i];
	  
	  if (r[kappa][kappa] <= 0.0)
	    {
	      zeros++;
	      kappa++;
	      appSP[kappa][kappa] = fpNorm (appB[kappa], n);
	      r[kappa][kappa] = appSP[kappa][kappa];
	    }
	  
	  kappa++;
	  
#ifdef DEBUG	  
	  Print_mat (B->coeff, kappamax, n);
	  Print_matf (appSP, expo, kappamax, kappamax);
	  printf("Step 8 is over, new kappa=%d\n",kappa);
#endif

	}
    };  
  

#ifdef VERBOSE
  tmp = 1.0;
  for (i = zeros+1; i<=d; i++)
    tmp*=r[i][i];
  tmp = sqrt (tmp);

  expo[zeros] = 0;
  for (i=zeros+1; i<=d; i++)
    expo[zeros] += expo[i];

  fprintf (stderr, "\nLoop iterations = %d \n", loops);
  fprintf (stderr, "Lovasz tests = %d \n", lovasz);
  fprintf (stderr, "Cputime is %d ms\n", (cputime()-start));
  if (zeros < d)
    {
#ifdef FORCE_LONGDOUBLE
      fprintf (stderr, "Vol(L) is %1.5Le * 2^%d\n", tmp, expo[zeros]);      
      fprintf (stderr, "||b_1|| is %1.5Le * 2^%d\n\n", 
	       sqrt ( r[zeros+1][zeros+1]), expo[zeros+1]);
#else
      fprintf (stderr, "Vol(L) is %E * 2^%d\n", tmp, expo[zeros]);      
      fprintf (stderr, "||b_1|| is %E * 2^%d\n\n", 
	       sqrt ( r[zeros+1][zeros+1]), expo[zeros+1]);
#endif
    }
#endif
  

  free (alpha);
  free (expo);
  mpz_clear (ztmp);
  clear_matrixf (mu, d);
  clear_matrixf (r, d);
  clear_matrixf (appB, d);
  clear_matrixf (appSP, d);
  free (s);
  free (appSPtmp);
}















/* ********************** */
/*  MAIN **************** */
/* ********************** */


#if 0

int
main (int argc, char *argv[])
{
  int d=0, i, j, n, bits, decal;
  double alpha;
  mat_ZZ B;
  char c=0;
#ifdef USING_MPFR
  size_t s;
  mpfr_t tmp;
  s = s;
#endif
  argc = argc;


  gmp_randinit_default (state);
  decal = 1;

  ctt = DELTA;
  halfplus = ETA;

  if (strcmp(argv[1],"-delta")==0) 
    {
      decal+=2;
      ctt = atof(argv[2]);
    }

  if (strcmp(argv[decal],"-eta")==0)
    {
      halfplus = atof(argv[decal+1]);
      decal+=2;
    }

  if (strcmp(argv[decal],"-delta")==0)
    {
      ctt = atof(argv[decal+1]);
      decal+=2;
    }

  onedothalfplus = 1.0+halfplus;

  if ((ctt >=1.0 ) || (ctt <=0.25) || 
      (halfplus <=0.5) || (halfplus>=sqrt(ctt)))
    {
      fprintf (stderr, "Incorrect parameters! You must choose\ndelta in (0.25,1) and eta in (0.5, sqrt(delta))\n"); 
      exit(1);
    }




  if ( strcmp(argv[decal], "h")==0) 
    {
      printf("Please have a look at the README file.\n");
      exit(1);
    }
  
  if (strcmp(argv[decal], "m")==0)
    {
      d = atoi(argv[decal+1]);
      n = atoi(argv[decal+2]);
      init_matrix (&B, d, n);

      /* read initial '[' */
      while (isspace(c = getchar ()) || c == '\n');
      if (c != '[')
        {
          fprintf (stderr, "Error: '[' expected\n");
          exit (1);
        }

      for (i=1; i<=d; i++)
	{
	  while (isspace(c = getchar ()) || c == '\n');
	  if (c != '[')
	    {
	      fprintf (stderr, "Error at row %d: '[' expected instead of %c\n", i, c);
	      exit (1);
	    }
	  for (j=1; j<=n; j++)
	    mpz_inp_str(B.coeff[i][j], stdin, 0);

	  while (isspace(c = getchar ()) || c=='\n');
	  if (c != ']')
	    {
	      fprintf (stderr, "Error: ']' expected at line %u\n", i);
	      exit (1);
	    }
	}
    }



  if (strcmp(argv[decal], "r")==0)
    {
      d = atoi(argv[decal+1]);
      bits = atoi(argv[decal+2]);
      init_matrix (&B, d, d+1);
      intrel (&B, bits, d);
    }      

  if (strcmp(argv[decal], "a")==0)
    {
      d = atoi(argv[decal+1]);
      alpha = atof(argv[decal+2]);
      init_matrix (&B, d, d);
      ajtai (&B, d, alpha);
    }  

  if (strcmp(argv[decal], "u")==0)
    {
      d = atoi(argv[decal+1]);
      bits = atoi(argv[decal+2]);
      init_matrix (&B, d, d);
      uniform (&B, bits, d); 
    }  

  if (strcmp(argv[decal], "n")==0)
    {
      d = atoi(argv[decal+1]);
      bits = atoi(argv[decal+2]);
      i = atoi(argv[decal+3]);
      init_matrix (&B, 2*d, 2*d);
      ntrulike (&B, bits, d, i); 
    }  


#ifdef USING_MPFR    
  if (strcmp(argv[decal], "k")==0)
    {
      d = atoi(argv[decal+1]);    
      init_matrix (&B, d, d+1);
      bits = atoi(argv[decal+2]);      

      mpfr_init2 (tmp, 2*bits);

      for (i=1; i<=d; i++)
	{
	  while (isspace(c) || c=='\n') c = getchar ();
	  s = mpfr_inp_str (tmp, stdin, 10, GMP_RNDN); 
	  if (s==0) fprintf( stderr, "error while reading the input, i est %d \n", i), abort();

          mpfr_mul_2ui (tmp, tmp, bits, GMP_RNDN);
	  mpfr_get_z (B.coeff[i][1], tmp, GMP_RNDN);
	}
      c = getchar();
      
      for (i=1; i<=d; i++)
	{
	  for(j=2; j<=i; j++)
	    mpz_set_ui ( B.coeff[i][j],0);
	  mpz_set_ui ( B.coeff[i][i+1],1);
	  for(j=i+2; j<=d+1; j++)
	    mpz_set_ui ( B.coeff[i][j],0);
	}
    }

  if (strcmp(argv[decal], "b")==0)
    {
      d = atoi(argv[decal+1]);
      mpfr_set_default_prec (1000);
      init_matrix (&B, d, d);
      GenerateBad (&B, d);       
    }  
#endif

    
  //  Print_mat (B.coeff, d, B.NumRows);

  LLL (&B);

  Print_mat (B.coeff, B.NumCols, B.NumRows);

  clear_matrix (&B, B.NumCols, B.NumRows);

#ifdef USING_MPFR
  if (strcmp(argv[decal], "k")==0) mpfr_clear (tmp);
#endif

  return 0;
}

#endif
