/*
  Copyright 2005 Damien Stehl.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  This program implements ideas from the paper "Floating-point LLL Revisited", 
  by Phong Nguyen and Damien Stehl, in the Proceedings of Eurocrypt'2005, 
  Springer-Verlag; and was partly inspired by Shoup's NTL library: 
  http://www.shoup.net/ntl/

  It is part of fplll-1.2, downloadable at:
  http://www.loria.fr/~stehle/fplll-1.2.tar.gz 

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <ctype.h>
#include <limits.h>
#include <gmp.h>

#ifndef NOT_USING_MPFR
#define USING_MPFR
#endif

#ifdef USING_MPFR
#include <mpfr.h>
#endif

#ifndef USE_DOUBLE
#ifndef USE_DPE
#ifndef USE_MPFR
#ifndef USE_LONGDOUBLE
#define USE_DPE
#endif
#endif
#endif
#endif

#ifdef USE_DPE
#ifdef FORCE_LONGDOUBLE
#define DPE_USE_DOUBLE
#endif
#endif

#if (LONG_MAX==2147483647L)
# define CPU_32
#define CPU_SIZE 32
#else
# define CPU_64
#define CPU_SIZE 64
#endif

#ifndef ETA
#define ETA 0.6
#endif
#ifndef DELTA
#define DELTA 0.8
#endif

gmp_randstate_t state;


#include "mat.h"
#include "heuristic.h"

FP_NR halfplus, onedothalfplus, ctt;

// #define VERBOSE
// #define TRIANGULAR
#define CANCELLATIONS
// #define DEBUG 

#ifdef TRIANGULAR
#ifndef SHIFT 
#define SHIFT 0
#endif
#endif

#ifdef DEBUG
#define LOOPS_BABAI 15
#endif

#ifdef VERBOSE
int
cputime ()
{
  struct rusage rus;

  getrusage (RUSAGE_SELF, &rus);
  return rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_usec / 1000;
}
#endif



/* computes the prec MSBs of the inner product of the approximate 
   vec1 and vec2.
   If too many bits are lost, then it returns 0, otherwise 1.
 */


int 
fpScalarProduct (FP_NR s, FP_NR *vec1, FP_NR *vec2, int n)
{
  int i, res;
  FP_NR tmp;
#ifdef CANCELLATIONS
  FP_NR tmp2;
#endif

  INIT (tmp);
#ifdef CANCELLATIONS
  INIT(tmp2);
#endif

  MUL (s, vec1[1], vec2[1]);

  for (i=2; i<=n; i++)
    {
      MUL (tmp, vec1[i], vec2[i]);      
      ADD (s, s, tmp);      
    }

#ifdef CANCELLATIONS
  fpNorm (tmp, vec1, n);
  fpNorm (tmp2, vec2, n);
  MUL (tmp, tmp, tmp2);
  DIV_2UI (tmp, tmp, 70);
  MUL (tmp2, s, s);

  if (CMP(tmp2, tmp)<=0)
    res = 0;
  else
#endif
    res = 1;

  CLEAR (tmp);
#ifdef CANCELLATIONS
  CLEAR (tmp2);
#endif
  return res;
} 


void 
fpNorm (FP_NR s, FP_NR *vec, int n)
{
  int i;
  FP_NR tmp;

  INIT (tmp);

  MUL (s, vec[1], vec[1]);

  for (i=2; i<=n; i++)
    {
      MUL (tmp, vec[i], vec[i]);
      ADD (s, s, tmp);      
    }
  
  CLEAR (tmp);
} 





/************************************/
/* ******************************** */
/* Babai's Nearest Plane algorithm  */
/* ******************************** */
/************************************/

/* Size-reduces b_kappa using mu_ij and r_ij for j<=i <kappa
updates B (kappa)
computes mu_kappaj, r_kappaj for j<=kappa, and s(kappa) 
The algorithm is the iterative Babai algorithm of the paper.
*/

void 
Babai (int kappa, mat_ZZ *B, mat_RR *mu, mat_RR *r, FP_NR *s, 
       mat_RR *appB, mat_RR *appSP, int a, int zeros, int kappamax, int n, 
       mpz_t ztmp, mpz_t X, FP_NR tmp, FP_NR rtmp)
{
  int i, j, k, test, aa, sg;
  signed long xx;
  mp_exp_t expo;
#ifdef DEBUG
  int loops=0;
#endif 


  if (a > zeros)
    aa = a;
  else
    aa = zeros+1;

#ifdef DEBUG
  printf("\nappSP: \n");
  Print_matf (appSP->coeff, kappamax, kappamax);
  printf("\nr: \n");
  Print_matf (r->coeff, kappamax, kappamax);
  printf("\n          STARTING BABAI WITH k=%d\n", kappa);
  printf("\nappB: \n");
  Print_matf (appB->coeff, kappamax, n);
  loops = 0;
  printf("\nmu: \n");
  Print_matf (mu->coeff, kappamax, kappamax);
  printf ("\n\na is %d, zeros is %d, aa is %d\n", a, zeros, aa);
  Print_mat (B->coeff, kappamax, n);
#endif

  do
    {
      test=0;

#ifdef DEBUG      
      if (loops++ > LOOPS_BABAI) 
	{	  
	  printf("INFINITE LOOP?\n");
	  exit(0);
	}
#endif 


      /* ************************************** */
      /* Step2: compute the GSO for stage kappa */
      /* ************************************** */

      for (j=aa; j<kappa; j++)
	{
	  if (IS_NAN(appSP->coeff[kappa][j])) 
	    {
	      if (!(fpScalarProduct (appSP->coeff[kappa][j], 
				    appB->coeff[kappa], 
				    appB->coeff[j], n)))
		{
		  ScalarProduct (ztmp, B->coeff[kappa], B->coeff[j], n);
		  SET_Z (appSP->coeff[kappa][j], ztmp);
		}
	    }

#ifdef DEBUG
	  if (loops<=LOOPS_BABAI)
	    {
	      printf("\n          j is %d\n", j);	  
	      printf("          appSP[%d][%d] is: ", kappa, j); 
	      OUT (appSP->coeff[kappa][j]); 
	      printf(", which approximates ");
	      ScalarProduct (ztmp, B->coeff[kappa], B->coeff[j], n);
	      printf("\n");
	      Print_mat (B->coeff, kappa, n);
	      mpz_out_str(stdout, 10, ztmp); 
	      printf("\n          Norm of B[%d]^2: ", j);
	      ScalarProduct (ztmp, B->coeff[j], B->coeff[j], n);
	      mpz_out_str(stdout, 10, ztmp);  
	      printf("\n          Norm of B[%d]^2: ",kappa);
	      ScalarProduct (ztmp, B->coeff[kappa], B->coeff[kappa], n);
	      mpz_out_str(stdout, 10, ztmp); 
	    }
#endif
	  
	  if (j > zeros+2)
	    {
	      MUL (tmp, mu->coeff[j][zeros+1], r->coeff[kappa][zeros+1]);
	      SUB (rtmp, appSP->coeff[kappa][j], tmp);
	      for (k=zeros+2; k<j-1; k++)
		{
		  MUL (tmp, mu->coeff[j][k], r->coeff[kappa][k]);
		  SUB (rtmp, rtmp, tmp);
		}
	      MUL (tmp, mu->coeff[j][j-1], r->coeff[kappa][j-1]);
	      SUB (r->coeff[kappa][j], rtmp, tmp);
	    }
	  else if (j==zeros+2)
	    {	      
	      MUL (tmp, mu->coeff[j][zeros+1], r->coeff[kappa][zeros+1]);
	      SUB (r->coeff[kappa][j], appSP->coeff[kappa][j], tmp);
	    }
	  else SET (r->coeff[kappa][j], appSP->coeff[kappa][j]);

	  DIV (mu->coeff[kappa][j], r->coeff[kappa][j], r->coeff[j][j]);
	}
      

      

#ifdef DEBUG
      if (loops <=LOOPS_BABAI)
	{
	  printf("\nmu :\n");
	  Print_matf (mu->coeff, kappa, kappa);
	  printf("\nr :\n");
	  Print_matf (r->coeff, kappa, kappa);
	}
#endif

      /* **************************** */
      /* Step3--5: compute the X_j's  */
      /* **************************** */
      
      for (j=kappa-1; j>zeros; j--)
	{
	  /* test of the relaxed size-reduction condition */
	  
	  ABS (tmp, mu->coeff[kappa][j]);

#ifdef DEBUG
	  if (loops<=LOOPS_BABAI)
	    {
	      printf( "tmp is : "); 
	      OUT(tmp); printf( "\n");
	    }
#endif

	  if ((CMP (tmp, halfplus)) > 0) 
	    {
	      test = 1; 
	      /* we consider separately the case X = +-1 */     
	      if ((CMP (tmp, onedothalfplus))<=0 )   
		{
#ifdef DEBUG
		  printf(" X is pm1\n");
#endif

		  sg = SGN ( mu->coeff[kappa][j]);
		  if ( sg >=0 )   /* in this case, X is 1 */
		    {
		      for (k=zeros+1; k<j; k++)
			SUB (mu->coeff[kappa][k], 
			     mu->coeff[kappa][k], 
			     mu->coeff[j][k]);
		      
		      for (i=1; i<=n; i++)    
			mpz_sub (B->coeff[kappa][i], 
				 B->coeff[kappa][i], 
				 B->coeff[j][i]);	  
		    }
		  
		  else          /* otherwise X is -1 */ 
		    {
		      for (k=zeros+1; k<j; k++)
			ADD (mu->coeff[kappa][k], 
			     mu->coeff[kappa][k], 
			     mu->coeff[j][k]);
		      
		      for (i=1; i<=n; i++)    
			mpz_add (B->coeff[kappa][i], 
				 B->coeff[kappa][i], 
				 B->coeff[j][i]);
		    }
		}
	      
	      else   /* we must have |X| >= 2 */
		{		  
		  RND (tmp, mu->coeff[kappa][j]);
		  
		  for (k=zeros+1; k<j; k++)
		    {
		      MUL (rtmp, tmp, mu->coeff[j][k]);
		      SUB (mu->coeff[kappa][k], mu->coeff[kappa][k], rtmp);
		    }
		  

#ifdef NO_DOUBLE
		  if (EXP(tmp) < CPU_SIZE-2)   
		    /* X is stored in a long signed int */		  
#else
		  if ((tmp[0] < (long double) 0x1p30)
		      &&(tmp[0] > (long double) -0x1p30))  
#endif
		      
		    {		      
		      GET_SI (xx, tmp);
		      
#ifdef DEBUG
		      if (loops<=LOOPS_BABAI)
			{
			  printf("          xx[%d] is %ld\n", j, xx);
			  printf("          and tmp was ");
			  OUT(tmp); printf("\n");
			}
#endif
		      
		      for (i=1; i<=n; i++)  
			{
			  mpz_mul_si (ztmp, B->coeff[j][i], xx);
			  mpz_sub (B->coeff[kappa][i], 
				   B->coeff[kappa][i], ztmp);
			}
		    }
		  
		  else
		    {
		      expo = GET_Z_EXP (X, tmp);
#ifdef DEBUG
		      printf("tmp is "); OUT (tmp);
		      printf("\nand expo is %ld, and X is ", expo);
		      mpz_out_str(stdout, 10, X); printf("\n");
#endif

		      if (expo <= 0)
			{
			  mpz_div_2exp (X, X, -expo);
			  expo = 0;

			  for (i=1; i<=n; i++)  
			    {
			      mpz_mul (ztmp, X, B->coeff[j][i]);
			      mpz_sub (B->coeff[kappa][i], 
				       B->coeff[kappa][i], ztmp);
			    }
			}
		      else
			{
			  for (i=1; i<=n; i++)  
			    {
			      mpz_mul (ztmp, X, B->coeff[j][i]);
			      mpz_mul_2exp (ztmp, ztmp, expo); 
			      mpz_sub (B->coeff[kappa][i], 
				       B->coeff[kappa][i], ztmp);
			    }
			}
		    }		  
		}
	    }
	}
      
      
      if (test)   /* Anything happened? */
	{
	  for (i=1 ; i<=n; i++)     /* update appB[kappa] */
	    SET_Z (appB->coeff[kappa][i], B->coeff[kappa][i]);	  
	  aa = zeros+1;
	  for (i=zeros+1; i<=kappa; i++) 
	    SET_NAN (appSP->coeff[kappa][i]);
	  for (i=kappa+1; i<=kappamax; i++) 
	    SET_NAN (appSP->coeff[i][kappa]);
	}

#ifdef DEBUG
      if (loops<=LOOPS_BABAI)
	{
	  printf("          test is %d\n", test);
	  printf("\nmu: \n");
	  Print_matf (mu->coeff, kappa, kappa);
	  printf("\nr: \n");
	  Print_matf (r->coeff, kappa, kappa);
	}
#endif

    }
  while (test);



  if (IS_NAN(appSP->coeff[kappa][kappa])) 
    {
      fpNorm (appSP->coeff[kappa][kappa], appB->coeff[kappa], n);
    }
  SET (s[zeros], appSP->coeff[kappa][kappa]);
  
  /* the last s[kappa-1]=r[kappa][kappa] is computed only if kappa increases */
  for (k=zeros+1; k<=kappa-2; k++)
    {
      MUL (tmp, mu->coeff[kappa][k], r->coeff[kappa][k]);
      SUB (s[k], s[k-1], tmp);
    }
  SET (r->coeff[kappa][kappa], s[kappa-1]);



#ifdef DEBUG
  printf("          Number of loops is %d\n", loops);
#endif

}





/* ****************** */
/* ****************** */
/* ****************** */
/* The LLL Algorithm  */
/* ****************** */
/* ****************** */
/* ****************** */

/* LLL-reduces the integer matrix B "in place" */


void 
LLL (mat_ZZ *B)
{
  int kappa, kappa2, d, n, i, j, zeros, kappamax;
  mat_RR mu, r, appB, appSP;
  FP_NR *s, *mutmp, *appBtmp, *appSPtmp;
  FP_NR tmp, rtmp;
  mpz_t ztmp, X;
  mpz_t *Btmp;
  int *alpha;
#ifdef VERBOSE
  int newkappa, start, lovasz, loops;
#endif
#ifdef FORCE_LONGDOUBLE
  long double tmp2, y;
#endif

  n = B->NumRows;
  d = B->NumCols;

  INIT (ctt);
  INIT (halfplus);

  SET_D (ctt, DELTA);
  SET_D (halfplus, ETA);

  alpha = (int *) malloc((d+1) * sizeof(int)); 
  INIT (tmp);
  INIT (rtmp);
  mpz_init (ztmp);
  mpz_init (X);

#ifdef DEBUG
  printf ("d = %d, n=%d\n", d, n);
#endif


#ifdef VERBOSE
  fprintf (stderr, 
	   "\nEntering LLL^2: LLL-reduction factors ( ");
  OUT2 (ctt); fprintf (stderr, ", ");
  OUT2 (halfplus); fprintf(stderr, ")\n");
#ifdef USE_MPFR
  fprintf (stderr, 
	   "Working precision set to %d bits\n", 
	   (int) mpfr_get_default_prec());
  if (mpfr_get_default_prec()<53) fprintf(stderr, "Be careful, your precision is particularly low.\n");
#else
#ifdef FORCE_LONGDOUBLE
  tmp2 = 1.0;
  for (j = 0; (y = tmp2 + 1.0) != tmp2; j++, tmp2 = 2.0 * tmp2);
  fprintf ( stderr, 
	    "Working precision set to %d bits.\n", j);
#else
  fprintf ( stderr, 
	    "Working precision set to 53 bits.\n");
#endif
#endif
  fprintf (stderr, 
	   "No efficiency nor correctness guarantee here.\nIf you want to be sure, LLL_check your output\nbasis or use the guaranteed version.\n\n");
#endif


  init_matrixf (&mu, d, d);
  init_matrixf (&r, d, d);
  
  init_matrixf (&appB, d, n);
  init_matrixf (&appSP, d, d);
  
  s = init_vecf (d+1);
  appSPtmp = init_vecf (d+1);
  
  for (i=0; i<=d; i++)
    for (j=0; j<=d; j++) 
      SET_NAN (appSP.coeff[i][j]);
  
  
  
  /* ************************** */
  /* Step1: Initialization Step */
  /* ************************** */     

#ifdef VERBOSE  
  start = cputime();
#endif
  

  for (i=1; i<=d; i++)
    for (j=1; j<=n; j++)
      SET_Z (appB.coeff[i][j], B->coeff[i][j]);


#ifdef DEBUG  
  Print_mat (B->coeff, d, n);
  Print_matf (appB.coeff, d, n);
  printf("Step 1 is over\n");
#endif  
  

  /* ********************************* */
  /* Step2: Initializing the main loop */
  /* ********************************* */   


#ifdef VERBOSE  
  newkappa = 1;
  loops = lovasz = 0;
#endif

  kappamax = 1;
  i = 1; 
  
  do
    fpNorm (appSP.coeff[i][i], appB.coeff[i], n); 
  while ((SGN (appSP.coeff[i][i]) == 0)&&(++i <=d));
  zeros = i-1; /* all vectors B[i] with i <= zeros are zero vectors */
  kappa = i+1;

  if (zeros < d) SET (r.coeff[zeros+1][zeros+1], appSP.coeff[zeros+1][zeros+1]);

  for (i=zeros+1; i<=d; i++)
    alpha[i]=1;
  
#ifdef DEBUG
  printf("Step 2 is over\n"); 
#endif  
  
  while (kappa <= d)
    {
      
      if (kappa>kappamax) kappamax++;
      
#ifdef VERBOSE
      loops++;
      if (kappa>newkappa)
	{
	  newkappa++;
	  fprintf (stderr, 
		   "Discovering vector k = %d, iterations = %d ", 
		   kappa, loops);
	  fprintf(stderr, "cputime is %d ms\n", (cputime()-start));
	}
#endif
      

#ifdef DEBUG
      printf("alpha= ");
      for (i=zeros+1; i<=d; i++) 
	printf("%d ",alpha[i]);
      printf("\n");
      printf("entering the while loop with k=%d\n", kappa);
      Print_mat (B->coeff, d, n);
#endif

    
      /* ********************************** */
      /* Step3: Call to the Babai algorithm */
      /* ********************************** */   

#ifdef TRIANGULAR   
      if (kappamax + SHIFT <= n){	
	Babai (kappa, B, &mu, &r, s, &appB, &appSP, 
	       alpha[kappa], zeros, kappamax, kappamax+SHIFT, 
	       ztmp, X, tmp, rtmp);
      }
      else {
	Babai (kappa, B, &mu, &r, s, &appB, &appSP, 
	       alpha[kappa], zeros, kappamax, n, ztmp,
	       X, tmp, rtmp);
      }
#else
      Babai (kappa, B, &mu, &r, s, &appB, &appSP, 
	     alpha[kappa], zeros, kappamax, n, 
	     ztmp, X, tmp, rtmp);
#endif

#ifdef DEBUG      
      printf("Step 3 is over\n");   
      Print_mat (B->coeff, kappamax, n);
      Print_matf (appB.coeff, kappamax, n);
      printf("\nappSP: \n");
      Print_matf (appSP.coeff, kappamax, kappamax);
      printf("\nr: \n");
      Print_matf (r.coeff, kappamax, kappamax);
#endif
      
      /* ************************************ */
      /* Step4: Success of Lovasz's condition */
      /* ************************************ */  
      /* ctt * r.coeff[kappa-1][kappa-1] <= s[kappa-2] ?? */
      
      MUL (tmp, r.coeff[kappa-1][kappa-1], ctt);
      
#ifdef DEBUG
      printf("s[%d] is ", kappa-2);
      OUT(s[kappa-2]); printf("\n");
#endif
      
#ifdef VERBOSE
      lovasz++;
#endif

      if( (CMP (tmp, s[kappa-2])) <=0 ) 
	{
	  alpha[kappa] = kappa;
	  MUL (tmp, mu.coeff[kappa][kappa-1], r.coeff[kappa][kappa-1]);
	  SUB (r.coeff[kappa][kappa], s[kappa-2], tmp);
	  kappa++;
	} 
      else
	{
	  
	  /* ******************************************* */
	  /* Step5: Find the right insertion index kappa */
          /* kappa2 remains the initial kappa            */
	  /* ******************************************* */  
	  
	  
	  kappa2 = kappa;
	  do
	    {
#ifdef VERBOSE
	      lovasz++;
#endif
	      kappa--;
	      MUL (tmp, r.coeff[kappa-1][kappa-1], ctt);
	    }
	  while ((kappa>=zeros+2) && ((CMP (s[kappa-2], tmp) <=0 )));
	  
#ifdef DEBUG
	  printf( "Index of insertion: %d \n", kappa);
	  printf("Step 5 is over\n");
	  printf("alpha= ");
	  for (i=1; i<=kappamax; i++) 
	    printf("%d ", alpha[i]);
	  printf("\n");
#endif

	  for (i=kappa; i<kappa2; i++)
	    if (kappa <= alpha[i]) alpha[i] = kappa;

	  for (i=kappa2; i>kappa; i--)
	    alpha[i] = alpha[i-1];
	  
	  for (i=kappa2+1; i<=kappamax; i++)
	    if (kappa < alpha[i]) alpha[i] = kappa;
	  
	  alpha[kappa] = kappa;
	  
#ifdef DEBUG
	  printf("alpha= ");
	  for (i=1; i<=kappamax; i++) 
	    printf("%d ", alpha[i]);
	  printf("\n");
#endif
	  
	  /* ****************************** */
	  /* Step6: Update the mu's and r's */
	  /* ****************************** */  

	  mutmp = mu.coeff[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    mu.coeff[i] = mu.coeff[i-1];
	  mu.coeff[kappa] = mutmp;

	  mutmp = r.coeff[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    r.coeff[i] = r.coeff[i-1];
	  r.coeff[kappa] = mutmp;

	  SET (r.coeff[kappa][kappa], s[kappa-1]);
	  
#ifdef DEBUG 
	  printf("Step 6 is over\n");
#endif
	  
	  /* ************************ */
	  /* Step7: Update B and appB */
	  /* ************************ */  	  
	  
	  Btmp = B->coeff[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    B->coeff[i] = B->coeff[i-1];
	  B->coeff[kappa] = Btmp;
	  
	  appBtmp = appB.coeff[kappa2];
	  for (i=kappa2; i>kappa; i--)
	    appB.coeff[i] = appB.coeff[i-1];
	  appB.coeff[kappa] = appBtmp;
	  
#ifdef DEBUG
	  Print_mat (B->coeff, d, n);
	  Print_matf (appB.coeff, d, n);
	  printf("Step 7 is over\n");
#endif
	  
	  /* *************************** */
	  /* Step8: Update appSP: tricky */
	  /* *************************** */  	 
	  
	  for (i=1; i<=kappa2; i++)
	    SET (appSPtmp[i], appSP.coeff[kappa2][i]);
	  
	  for (i=kappa2+1; i<=kappamax; i++)
	    SET (appSPtmp[i], appSP.coeff[i][kappa2]);
	  
	  for (i=kappa2; i>kappa; i--)
	    {
	      for (j=1; j<kappa; j++)
		SET (appSP.coeff[i][j], appSP.coeff[i-1][j]);
	      
	      SET (appSP.coeff[i][kappa], appSPtmp[i-1]);
	      
	      for (j=kappa+1; j<=i; j++)
		SET (appSP.coeff[i][j], appSP.coeff[i-1][j-1]);
	      
	      for (j=kappa2+1; j<=kappamax; j++)
		SET (appSP.coeff[j][i], appSP.coeff[j][i-1]);     
	    }
	  
	  for (i=1; i<kappa; i++)
	    SET (appSP.coeff[kappa][i], appSPtmp[i]);
	  
	  SET (appSP.coeff[kappa][kappa], appSPtmp[kappa2]);
	  
	  for (i=kappa2+1; i<=kappamax; i++)
	    SET (appSP.coeff[i][kappa], appSPtmp[i]);
	  

	  if (SGN(r.coeff[kappa][kappa]) <= 0)
	    {
	      zeros++;
	      kappa++;
	      fpNorm (appSP.coeff[kappa][kappa], appB.coeff[kappa], n);
	      SET (r.coeff[kappa][kappa], appSP.coeff[kappa][kappa]);
	    }
	  
	  kappa++;
	  
#ifdef DEBUG	  
	  Print_matf (appB.coeff, kappamax, n);
	  Print_matf (appSP.coeff, kappamax, kappamax);
	  printf("Step 8 is over, new kappa=%d\n",kappa);
#endif
	  
	}
    };  
  
#ifdef VERBOSE
  SET_D (tmp, 1.0);
  for (i = zeros+1; i<=d; i++)
    MUL (tmp, tmp, r.coeff[i][i]);
  SQRT (tmp, tmp);

  fprintf (stderr, "\nLoop iterations = %d \n", loops);
  fprintf (stderr, "Lovasz tests = %d \n", lovasz);
  fprintf (stderr, "Cputime is %d ms\n", (cputime()-start));
  if (zeros < d)
    {
      fprintf (stderr, "Vol(L) is "); OUT2(tmp); fprintf(stderr, "\n");      
      SQRT (tmp, r.coeff[zeros+1][zeros+1]);
      fprintf (stderr, "||b_1|| is "); 
      OUT2(tmp); fprintf(stderr, "\n\n");
    }
#endif
  
  
  
  free (alpha);
  
  clear_matrixf (&mu, d, d);
  clear_matrixf (&r, d, d);
  clear_matrixf (&appB, d, n);
  clear_matrixf (&appSP, d, d);
  
  CLEAR (tmp);
  CLEAR (rtmp);

  CLEAR (ctt);
  CLEAR (halfplus);

  mpz_clear (ztmp);
  mpz_clear (X);

  clear_vecf (s, d+1);
  clear_vecf (appSPtmp, d+1);
}


#if 0



/* ********************** */
/*  MAIN **************** */
/* ********************** */

int
main (int argc, char *argv[])
{
  int d=0, i, j, n, bits, decal;
  double alpha, delta, eta;
  mat_ZZ B;
  char c=0;
#ifdef USING_MPFR
  mpfr_t tmp;
  size_t s;
#endif
#ifdef USE_MPFR
  double rho;
  int wantedprec=0, goodprec;
#endif
  argc = argc;
#ifdef USING_MPFR
  s = s;
#endif

  INIT (halfplus);
  INIT (onedothalfplus);
  INIT (ctt);
  
  gmp_randinit_default (state);
  decal = 1;

  delta = DELTA;
  eta = ETA;

  if (strcmp(argv[1],"-delta")==0) 
    {
      decal+=2;
      delta = atof(argv[2]);
    }

  if (strcmp(argv[decal],"-eta")==0)
    {
      eta = atof(argv[decal+1]);
      decal+=2;
    }

  if (strcmp(argv[decal],"-delta")==0)
    {
      delta = atof(argv[decal+1]);
      decal+=2;
    }


  SET_D (halfplus, eta) ;
  SET_D (onedothalfplus, 1.0+eta);
  SET_D (ctt, delta);  

  if ((delta >=1.0 ) || (delta <=0.25) || (eta <=0.5) || (eta>=sqrt(delta)))
    {
      fprintf (stderr, "Incorrect parameters! You must choose\ndelta in (0.25,1) and eta in (0.5, sqrt(delta))\n"); 
      exit(1);
    }



  if ( strcmp(argv[decal], "h")==0) 
    {
      printf("Please have a look at the README file.\n");
      exit(1);
    }
  
  if (strcmp(argv[decal], "m")==0)
    {
      d = atoi(argv[decal+1]);
      n = atoi(argv[decal+2]);
      init_matrix (&B, d, n);

#ifdef USE_MPFR
      if (argc==decal+4) wantedprec = atoi(argv[decal+3]);
#endif

      /* read initial '[' */
      while (isspace(c = getchar ()) || c == '\n');
      if (c != '[')
        {
          fprintf (stderr, "Error: '[' expected\n");
          exit (1);
        }

      for (i=1; i<=d; i++)
	{
	  while (isspace(c = getchar ()) || c == '\n');
	  if (c != '[')
	    {
	      fprintf (stderr, "Error at row %d: '[' expected instead of %c\n", i, c);
	      exit (1);
	    }
	  for (j=1; j<=n; j++)
	    mpz_inp_str(B.coeff[i][j], stdin, 0);

	  while (isspace(c = getchar ()) || c=='\n');
	  if (c != ']')
	    {
	      fprintf (stderr, "Error: ']' expected at line %u\n", i);
	      exit (1);
	    }
	}
    }

  
  if (strcmp(argv[decal], "r")==0)
    {
      d = atoi(argv[decal+1]);
      bits = atoi(argv[decal+2]);

#ifdef USE_MPFR
      if (argc==decal+4) wantedprec = atoi(argv[decal+3]);
#endif

      init_matrix (&B, d, d+1);
      intrel (&B, bits, d);
    }      
 
  if (strcmp(argv[decal], "n")==0)
    {
      d = atoi(argv[decal+1]);
      bits = atoi(argv[decal+2]);
      i = atoi(argv[decal+3]);
#ifdef USE_MPFR
      if (argc==decal+5) wantedprec = atoi(argv[decal+4]);
#endif
      init_matrix (&B, 2*d, 2*d);
      ntrulike (&B, bits, d, i); 
    }  


 
  if (strcmp(argv[decal], "a")==0)
    {
      d = atoi(argv[decal+1]);
      alpha = atof(argv[decal+2]);

#ifdef USE_MPFR
      if (argc==decal+4) wantedprec = atoi(argv[decal+3]);
#endif

      init_matrix (&B, d, d);
      ajtai (&B, d, alpha);
    }  

  if (strcmp(argv[decal], "u")==0)
    {
      d = atoi(argv[decal+1]);
      bits = atoi(argv[decal+2]);

#ifdef USE_MPFR
      if (argc==decal+4) wantedprec = atoi(argv[decal+3]);
#endif

      init_matrix (&B, d, d);
      uniform (&B, bits, d); 
    }  
    
#ifdef USING_MPFR
  if (strcmp(argv[decal], "b")==0)
    {
      d = atoi(argv[decal+1]);

#ifdef USE_MPFR
      if (argc==decal+3) wantedprec = atoi(argv[decal+2]);
#endif	
      mpfr_set_default_prec (1000);
      init_matrix (&B, d, d);
      GenerateBad (&B, d);       
    }  


  if (strcmp(argv[decal], "k")==0)
    {
      d = atoi(argv[decal+1]);    
      init_matrix (&B, d, d+1);
      bits = atoi(argv[decal+2]);      

      mpfr_init2 (tmp, 2*bits);

#ifdef USE_MPFR
      if (argc==decal+4) wantedprec = atoi(argv[decal+3]);
#endif	

      for (i=1; i<=d; i++)
	{
	  while (isspace(c) || c=='\n') c = getchar ();
	  s = mpfr_inp_str (tmp, stdin, 10, GMP_RNDN); 
	  if (s==0) fprintf( stderr, "error while reading the input, i est %d \n", i), abort();

          mpfr_mul_2ui (tmp, tmp, bits, GMP_RNDN);
	  mpfr_get_z (B.coeff[i][1], tmp, GMP_RNDN);
	}
      c = getchar();
      
      for (i=1; i<=d; i++)
	{
	  for(j=2; j<=i; j++)
	    mpz_set_ui ( B.coeff[i][j],0);
	  mpz_set_ui ( B.coeff[i][i+1],1);
	  for(j=i+2; j<=d+1; j++)
	    mpz_set_ui ( B.coeff[i][j],0);
	}
    }
#endif
    
  //Print_mat (B.coeff, d, B.NumRows);

#ifdef USE_MPFR
  rho = (1.0 + eta) * (1.0+eta) / (delta - eta*eta);
  goodprec = (unsigned int) (7.0 + 0.2*d + d* log(rho)/ log(2.0) 
			     +  2.0 * log ( (double) d )  
			     - log( (eta-0.5)*(1.0-delta) ) / log(2));

  if (wantedprec == 0) wantedprec = goodprec;
  mpfr_set_default_prec (wantedprec);
#endif

  LLL(&B);

  Print_mat (B.coeff, d, B.NumRows);


  clear_matrix (&B, B.NumCols, B.NumRows);

#ifdef USING_MPFR
  if (strcmp(argv[decal], "k")==0) mpfr_clear (tmp);
#endif

  CLEAR (halfplus);
  CLEAR (onedothalfplus);
  CLEAR (ctt);

  return 0;
}

#endif
